from fontTools.ttLib import TTFont
from .element import Text, Space
from .utils import Point


def build_title(
    text: str,
    x_end: float,
    font: TTFont,
) -> Text:
    space = Space(
        top_left=Point(x=0, y=0),
        bottom_right=Point(x=x_end, y=0),
    )

    title = Text(
        text=text,
        font_size=26,
        pos=Point(x=0.5, y=1),
        space=space,
        x_align="center",
        y_align="top",
        font=font,
    )

    return title



import os
from typing import Type, TypeVar
from newsflash.widgets.layout.base import Layout
from newsflash.widgets.base import Widget
from newsflash.widgets import Notifications, Columns

from django.http import HttpRequest
from django.core.management import execute_from_command_line
from django.core.wsgi import get_wsgi_application
from django.core.handlers.wsgi import WSGIHandler


T = TypeVar("T")


class App:
    layout: Layout
    navbar: bool = False

    def __init__(self) -> None:
        self.layout = self.compose()
        self.layout.main_element = True

    def compose(self) -> Layout:
        return Columns(children=[])

    def query_one(self, id: str, type: Type[T]) -> T | None:
        notifications = Notifications()
        if id == "notifications" and isinstance(notifications, type):
            return notifications

        return self.layout.query_one(id, type)

    def query_all(self, type: Type[T]) -> list[T]:
        return self.layout.query_all(type=type, widgets=[])

    def query_type(self, id: str) -> Type[Widget] | None:
        return self.layout.query_type_one(id)

    def render(self, request: HttpRequest) -> str:
        return self.layout.render(request)

    @classmethod
    def run(cls):
        from newsflash.web.app.urls import set_urlpatterns

        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "newsflash.web.settings")
        set_urlpatterns(cls)
        execute_from_command_line(["manage.py", "runserver"])

    @classmethod
    def get_wsgi_application(cls) -> WSGIHandler:
        from newsflash.web.app.urls import set_urlpatterns

        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "newsflash.web.settings")
        set_urlpatterns(cls)
        application = get_wsgi_application()
        return application



from typing import Callable
from inspect import signature

from .models import Callback, WidgetIO
from .utils import process_callback_arg


def construct_callback(
    callback_fn: Callable,
    endpoint_name: str,
    trigger_event: str,
) -> Callback:
    sig = signature(callback_fn)
    input_ids: list[str] = []
    target_ids: list[str] = []

    for param in sig.parameters:
        if param == "self" or param == "selected":
            continue

        _, widget_id, widget_io = process_callback_arg(
            callback_fn=callback_fn,
            parameter=param
        )

        if widget_io == WidgetIO.INPUT or widget_io == WidgetIO.BOTH:
            input_ids.append(widget_id + "-input")
        
        if widget_io == WidgetIO.OUTPUT or widget_io == WidgetIO.BOTH:
            target_ids.append(widget_id)

    return Callback(
        endpoint_name=endpoint_name,
        trigger_event=trigger_event,
        inputs=input_ids,
        targets=target_ids,
        target_wrapper_ids=", ".join([f"#{target}-wrapper" for target in target_ids]),
    )



from .base import Layout
from ..base import Widget


class Columns(Layout):
    template_name: str = "layout/columns"
    grow: bool = False


def build_columns(*children: Widget, span: int = 1) -> Columns:
    return Columns(children=list(children), span=span)



from typing import Callable
from django.core.cache import cache
from functools import wraps
import json
import hashlib


def cache_result(timeout: int = 60 * 60 * 24):
    def decorator(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            key = (
                f"{func.__module__}.{func.__name__}:"
                + hashlib.sha256(
                    json.dumps({"args": args, "kwargs": kwargs}).encode()
                ).hexdigest()
            )

            if (result := cache.get(key)) is not None:
                return result

            result = func(*args, **kwargs)
            cache.set(key, result, timeout)
            return result

        return wrapper

    return decorator



from typing import Callable, Literal
from inspect import signature
import json

from django.http import HttpRequest

from newsflash.widgets.base import Widget
from newsflash.widgets.control.select import Select
from newsflash.widgets.chart.base import Chart
from .utils import process_callback_arg
from .models import WidgetIO


# TODO: find better location (and deduplicate)
type ChartDimension = dict[Literal["width"] | Literal["height"], float]
type ChartDimensions = dict[str, ChartDimension]


def get_callback_in_and_outputs(
    callback_fn: Callable,
    additional_inputs: dict[str, str],
    chart_dimensions_dict: ChartDimensions,
) -> tuple[dict[str, Widget], dict[str, Widget]]:
    sig = signature(callback_fn)

    callback_inputs: dict[str, Widget] = {}
    callback_outputs: dict[str, Widget] = {}
    for param in sig.parameters:
        if param == "self" or param == "selected":
            continue

        widget_type, widget_id, widget_io = process_callback_arg(callback_fn, param)

        if issubclass(widget_type, Select):
            widget = widget_type(selected=additional_inputs[widget_id])
        elif issubclass(widget_type, Chart):
            widget = widget_type(
                width_in_px=chart_dimensions_dict[widget_id]["width"],
                height_in_px=chart_dimensions_dict[widget_id]["height"],
                swap_oob=True,
            )
        else:
            widget = widget_type()

        callback_inputs[param] = widget
        if widget_io == WidgetIO.OUTPUT or widget_io == WidgetIO.BOTH:
            widget.swap_oob = True
            callback_outputs[param] = widget

    return callback_inputs, callback_outputs


def render_callback_outputs(
    callback_outputs: dict[str, Widget], request: HttpRequest
) -> bytes:
    result: str = ""

    for callback_output in callback_outputs.values():
        if callback_output._cancel_update:
            continue
        if isinstance(callback_output, Chart):
            result += "\n" + callback_output.render_chart(
                request, authenticated=request.user.is_authenticated
            )
        else:
            result += "\n" + callback_output.render(request)

    return result.encode()


def parse_chart_dimensions(request: HttpRequest) -> ChartDimensions:
    dimensions = request.POST["dimensions"]
    assert isinstance(dimensions, str)
    chart_dimensions = [
        {k.removesuffix("-container"): v for k, v in chart.items()}
        for chart in json.loads(dimensions)
    ]

    chart_dimensions_dict: ChartDimensions = {}
    for chart in chart_dimensions:
        chart_dimensions_dict.update(chart)

    return chart_dimensions_dict


def parse_additional_inputs(request: HttpRequest, trigger: str) -> dict[str, str]:
    additional_inputs: dict[str, str] = {}

    for k in request.POST:
        if k.endswith("-value") and k.removesuffix("-value") != trigger:
            _value = request.POST[k]
            assert isinstance(_value, str)
            additional_inputs[k.removesuffix("-value")] = _value

    return additional_inputs


def parse_request_inputs(
    request: HttpRequest,
) -> tuple[str, dict[str, str], ChartDimensions]:
    trigger: str = request.headers["Hx-Trigger"]  # type: ignore
    additional_inputs = parse_additional_inputs(request, trigger)
    chart_dimensions_dict = parse_chart_dimensions(request)
    return trigger, additional_inputs, chart_dimensions_dict




from enum import StrEnum
from pydantic import BaseModel


class Callback(BaseModel):
    endpoint_name: str
    trigger_event: str
    inputs: list[str]
    targets: list[str]
    target_wrapper_ids: str


class WidgetIO(StrEnum):
    INPUT = "input"
    OUTPUT = "output"
    BOTH = "input+output"



from django.http import HttpRequest
from django.utils.text import slugify

from pydantic import BaseModel, model_validator
from ..base import TextWidget
from typing import Literal, Self


type TextAlign = Literal["left"] | Literal["center"] | Literal["right"]


class TitleContext(BaseModel):
    id: str
    title: str
    text_size: str
    level: int
    navbar_text_size: str
    navbar_margin_y: str | None
    navbar_margin_start: str | None
    align: TextAlign


def map_title_level_to_title_text_size(title_level: int) -> str:
    match title_level:
        case 1:
            return "4xl"
        case 2:
            return "3xl"
        case 3:
            return "2xl"
        case _:
            raise ValueError("invalid title level")
        

def map_title_level_to_navbar_text_size(title_level: int) -> str:
    match title_level:
        case 1:
            return "xl"
        case 2:
            return "xl"
        case 3:
            return "lg"
        case _:
            raise ValueError("invalid title level")
        

def map_title_level_to_margin_y(title_level: int) -> str | None:
    match title_level:
        case 1:
            return "3"
        case 2:
            return "1"
        case 3:
            return None
        case _:
            raise ValueError("invalid title level")
        

def map_title_level_to_margin_start(title_level: int) -> str | None:
    match title_level:
        case 1:
            return None
        case 2:
            return "2"
        case 3:
            return "4"
        case _:
            raise ValueError("invalid title level")


class Title(TextWidget):
    template_name: str = "text/title"
    title: str
    text_size: str | None = None
    navbar_text_size: str | None = None
    align: TextAlign = "left"
    level: int = 1

    @model_validator(mode="after")
    def assert_text_size(self) -> Self:
        if self.text_size is None:
            self.text_size = map_title_level_to_title_text_size(self.level)
        return self
    
    @model_validator(mode="after")
    def assert_navbar_text_size(self) -> Self:
        if self.navbar_text_size is None:
            self.navbar_text_size = map_title_level_to_navbar_text_size(self.level)
        return self
    
    @model_validator(mode="after")
    def assert_id(self) -> Self:
        if self.id is None:
            self.id = slugify(self.title)
        return self

    def update_title(self, new_title: str) -> None:
        self.title = new_title

    def _build(self, request: HttpRequest) -> TitleContext:
        assert self.id is not None
        assert self.text_size is not None
        assert self.navbar_text_size is not None

        return TitleContext(
            id=self.id, 
            title=self.title, 
            text_size=self.text_size,
            level=self.level,
            navbar_text_size=self.navbar_text_size,
            navbar_margin_y=map_title_level_to_margin_y(self.level),
            navbar_margin_start=map_title_level_to_margin_start(self.level),
            align=self.align,
        )


class SubTitle(Title):
    level: int = 2


class SubSubTitle(Title):
    level: int = 3



from inspect import signature, getsourcefile
from runpy import run_path
from typing import Callable, Type, get_args, ForwardRef

from .models import WidgetIO
from newsflash.widgets.base import Widget, ChartWidget, ControlWidget, TextWidget


def resolve_forward_ref(
    widget_type: Type[Widget], callback_fn: Callable
) -> Type[Widget]:
    if isinstance(widget_type, ForwardRef):
        source_file = getsourcefile(callback_fn)
        assert source_file is not None
        file_globals = run_path(source_file)
        widget_type = file_globals[widget_type.__forward_arg__]

    return widget_type


def process_callback_arg(
    callback_fn: Callable, parameter: str
) -> tuple[Type[Widget], str, WidgetIO]:
    sig = signature(callback_fn)
    annotation = sig.parameters[parameter].annotation
    args = get_args(annotation)

    widget_type = args[0]
    widget_type = resolve_forward_ref(widget_type, callback_fn)
    widget_id = args[1]

    try:
        widget_io = args[2]
    except IndexError:
        if issubclass(widget_type, ControlWidget):
            widget_io = WidgetIO.INPUT
        elif issubclass(widget_type, ChartWidget) or issubclass(widget_type, TextWidget):
            widget_io = WidgetIO.OUTPUT
        else:
            raise ValueError(f"unexpected widget_type: {widget_type}")

    return widget_type, widget_id, widget_io



from .title import Title, SubTitle, SubSubTitle
from .paragraph import Paragraph
from .notifications import Notifications


__all__ = [
    "Title",
    "SubTitle",
    "SubSubTitle",
    "Paragraph",
    "Notifications",
]



from typing import Callable, Type, Literal
import json

from django.http import HttpRequest, HttpResponse
from django.shortcuts import render

from newsflash import App
from newsflash.widgets.chart.base import Chart
from newsflash.widgets.control.select import Select
from newsflash.widgets import Button, Title
from newsflash.callback.caller import (
    get_callback_in_and_outputs, 
    render_callback_outputs,
    parse_request_inputs,
)


type ChartDimension = dict[Literal["width"] | Literal["height"], float]
type ChartDimensions = dict[str, ChartDimension]


def build_main_view(app: Type[App]) -> Callable:
    def main(request: HttpRequest, page_num: int = 1) -> HttpResponse:
        _app = app()
        titles = _app.query_all(type=Title)
        titles = [title._build(request=request) for title in titles]

        return render(
            request,
            "app/page.html",
            context={
                "content": _app.render(request),
                "page_num": page_num,
                "navbar": _app.navbar,
                "titles": titles
            },
        )

    return main


def build_button_view(app: Type[App]) -> Callable:
    def click(request: HttpRequest) -> HttpResponse:
        trigger, additional_inputs, chart_dimensions_dict = parse_request_inputs(
            request
        )

        _app = app()
        button_element = _app.query_one(trigger, Button)
        assert button_element is not None

        callback_inputs, callback_outputs = get_callback_in_and_outputs(
            button_element.on_click, additional_inputs, chart_dimensions_dict
        )

        button_element.on_click(**callback_inputs)
        return HttpResponse(render_callback_outputs(callback_outputs, request))

    return click


def build_select_view(app: Type[App]) -> Callable:
    def select(request: HttpRequest) -> HttpResponse:
        trigger, additional_inputs, chart_dimensions_dict = parse_request_inputs(
            request
        )
        trigger = trigger.removesuffix("-input")
        value = request.POST[f"{trigger}-value"]

        _app = app()
        select_element = _app.query_one(trigger, Select)
        assert select_element is not None

        callback_inputs, callback_outputs = get_callback_in_and_outputs(
            select_element.on_input, additional_inputs, chart_dimensions_dict
        )

        value_type = type(select_element.selected)
        select_element.selected = value_type(value)
        select_element.on_input(**callback_inputs)

        return HttpResponse(render_callback_outputs(callback_outputs, request))

    return select


def build_chart_view(app: Type[App]) -> Callable:
    def chart(request: HttpRequest) -> HttpResponse:
        trigger, additional_inputs, chart_dimensions_dict = parse_request_inputs(
            request
        )
        trigger = trigger.removesuffix("-wrapper")

        _app = app()
        chart_element = _app.query_one(trigger, Chart)
        assert chart_element is not None

        chart_element.width_in_px = chart_dimensions_dict[trigger]["width"]
        chart_element.height_in_px = chart_dimensions_dict[trigger]["height"]

        callback_inputs, callback_outputs = get_callback_in_and_outputs(
            chart_element.on_load, additional_inputs, chart_dimensions_dict
        )

        callback_outputs[trigger] = chart_element
        chart_element.on_load(**callback_inputs)

        return HttpResponse(render_callback_outputs(callback_outputs, request))

    return chart



from math import floor
from django.http import HttpRequest
from django.contrib.humanize.templatetags.humanize import intcomma
from .base import Chart
from pydantic import BaseModel
from .utils import Point, Space
from fontTools.ttLib import TTFont
from pathlib import Path

from .element import Text, ElementCollection
from .element import Path as PathElement
from .axes import build_x_axis, build_y_axis, build_y_grid_lines, build_y_ticks
from .title import build_title


def build_lines(
    xs: list[float],
    ys: list[float],
    multiplier: float,
    max_y: float,
    space: Space,
) -> ElementCollection:
    elements = ElementCollection()

    ys = [y / multiplier for y in ys]
    normalized_ys = [y / max_y for y in ys]
    min_x, max_x = xs[0], xs[-1]
    normalized_xs = [(x - min_x) / (max_x - min_x) for x in xs]

    points: list[Point] = []
    for idx, x in enumerate(normalized_xs):
        points.append(Point(x=x, y=normalized_ys[idx]))

    elements.append(
        PathElement(
            points=points,
            space=space,
            classes=["stroke-line-light", "dark:stroke-line-dark", "line-path"],
        )
    )

    return elements


class LineChartContext(BaseModel):
    id: str
    width: float
    height: float
    elements: list[str]
    swap_oob: bool = False


class LineChart(Chart):
    
    xs: list[float] = []
    ys: list[float] = []

    def set_points(self, xs: list[float], ys: list[float]) -> None:
        self.xs = xs
        self.ys = ys

    def _build_chart(self, request: HttpRequest, id: str) -> LineChartContext:
        font = TTFont(
            Path(__file__).resolve().parent.parent.parent
            / "fonts"
            / "noto-serif"
            / "NotoSerif.ttf"
        )

        elements = ElementCollection()

        if self.title is not None:
            title = build_title(
                text=self.title,
                x_end=self.width_in_px,
                font=font,
            )

            elements.append(title)
            title_height = title.get_height()
        else:
            title_height = 0

        x_axis_font_size = 16
        x_axis_height = x_axis_font_size * 1.5

        y_ticks, y_labels, multiplier = build_y_ticks(
            number_of_ticks=4,
            max_y_value=max(self.ys),
        )

        if multiplier != 1:
            multiplier_text = Text(
                text=f"x {intcomma(multiplier)}",
                pos=Point(x=0, y=title_height),
                y_align="bottom",
                font_size=x_axis_font_size,
                font=font,
            )
            elements.append(multiplier_text)

        y_axis = build_y_axis(
            y_start=title_height * 2,
            y_end=self.height_in_px - x_axis_height,
            x_start=0,
            bottom_y_value=0.0,
            top_y_value=max(y_ticks),
            ticks=y_ticks,
            labels=y_labels,
            font=font,
        )

        bar_width = ((self.width_in_px - y_axis.get_width()) / len(self.ys)) / 2
        chart_start_x = y_axis.get_width() + x_axis_font_size

        ticks_step = floor(len(self.xs) / 8)

        x_axis = build_x_axis(
            x_start=chart_start_x + bar_width,
            x_end=self.width_in_px - bar_width,
            y_end=self.height_in_px,
            start_x_value=self.xs[0],
            end_x_value=self.xs[-1],
            ticks=list(range(int(self.xs[0]), int(self.xs[-1]) + 1, ticks_step)),
            font_size=x_axis_font_size,
            font=font,
        )

        y_grid_lines = build_y_grid_lines(
            bottom_y_value=0.0,
            top_y_value=max(y_ticks),
            ticks=y_ticks,
            space=Space(
                top_left=Point(x=chart_start_x + bar_width, y=title_height * 2),
                bottom_right=Point(
                    x=self.width_in_px - bar_width, y=self.height_in_px - x_axis_height
                ),
            ),
        )

        bars = build_lines(
            xs=self.xs,
            ys=self.ys,
            multiplier=multiplier,
            max_y=max(y_ticks),
            space=Space(
                top_left=Point(x=chart_start_x + bar_width, y=title_height * 2),
                bottom_right=Point(
                    x=self.width_in_px - bar_width, y=self.height_in_px - x_axis_height
                ),
            ),
        )

        elements.extend(y_axis)
        elements.extend(x_axis)
        elements.extend(y_grid_lines)
        elements.extend(bars)

        assert self.id is not None
        return LineChartContext(
            id=self.id,
            width=self.width_in_px,
            height=self.height_in_px,
            elements=elements.render(),
            swap_oob=self.swap_oob,
        )



import os
from django.core.management import execute_from_command_line


def main():
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "newsflash.web.settings")
    execute_from_command_line(["manage.py", "migrate"])
    execute_from_command_line(["manage.py", "createcachetable"])
    execute_from_command_line(["manage.py", "collectstatic"])
    execute_from_command_line(["manage.py", "createsuperuser"])



from typing import Literal, Self
from pydantic import BaseModel, ConfigDict
from newsflash.fonts import get_text_width, get_text_height
from .utils import chart_to_svg_coordinates, Point, Space
from fontTools.ttLib import TTFont
from django.template.loader import render_to_string


class Element(BaseModel):
    template_name: str = ""
    space: Space | None = None
    classes: list[str] = []

    def prerender(self) -> None:
        pass

    def render(self) -> str:
        if self.space is not None:
            self.transform_to_space(self.space)
        self.prerender()
        return render_to_string(
            f"app/widgets/chart/elements/{self.template_name}.html",
            context=self.model_dump(),
        )

    def get_width(self) -> float: ...

    def transform_to_space(self, space: Space) -> Self: ...


class Rectangle(Element):
    template_name: str = "rect"
    top_left: Point
    bottom_right: Point
    shape: Point | None = None
    fill_color: str = "lightblue"
    rounded: bool = False
    rx: float = 0
    ry: float = 0

    def transform_to_space(self, space: Space) -> Self:
        self.top_left = chart_to_svg_coordinates(space, self.top_left)
        self.bottom_right = chart_to_svg_coordinates(space, self.bottom_right)
        self.shape = Point(
            x=self.bottom_right.x - self.top_left.x,
            y=self.bottom_right.y - self.top_left.y,
        )
        width = self.bottom_right.x - self.top_left.x
        if self.rounded:
            self.rx = min(width * 0.05, 5)
            self.ry = min(width * 0.05, 5)
        return self

    def get_width(self) -> float:
        return self.bottom_right.x - self.top_left.x


def rect_from_bottom_center(
    bottom_center: Point,
    width: float,
    height: float,
    space: Space | None = None,
    rounded: bool = False,
) -> Rectangle:
    return Rectangle(
        top_left=Point(x=bottom_center.x - (width / 2), y=bottom_center.y + height),
        bottom_right=Point(x=bottom_center.x + (width / 2), y=bottom_center.y),
        space=space,
        rounded=rounded,
        classes=["fill-bar-light", "dark:fill-bar-dark", "bar"],
    )


class Line(Element):
    template_name: str = "line"
    from_pos: Point
    to_pos: Point
    path_length: float = 100
    stroke_color: str = "black"
    stroke_width: float = 5.0
    stroke_linecap: str = "round"

    def transform_to_space(self, space: Space) -> Self:
        self.from_pos = chart_to_svg_coordinates(space, self.from_pos)
        self.to_pos = chart_to_svg_coordinates(space, self.to_pos)
        return self

    def get_width(self) -> float:
        return abs(self.to_pos.x - self.from_pos.x)


class Path(Element):
    template_name: str = "path"
    points: list[Point]
    path_length: float = 100
    stroke_color: str = "black"
    stroke_width: float = 5.0
    stroke_linecap: str = "round"
    fill_color: str = "transparent"

    def transform_to_space(self, space: Space) -> Self:
        transformed_points: list[Point] = []
        for point in self.points:
            point = chart_to_svg_coordinates(space, point)
            transformed_points.append(point)

        self.points = transformed_points
        return self

    def get_width(self) -> float: ...


type YAlign = Literal["top"] | Literal["center"] | Literal["bottom"]
type XAlign = Literal["left"] | Literal["center"] | Literal["right"]


class Text(Element):
    template_name: str = "text"
    text: str
    font_size: int
    pos: Point
    x_align: XAlign = "left"
    y_align: YAlign = "center"
    font: TTFont
    classes: list[str] = ["fill-text-light", "dark:fill-text-dark"]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def prerender(self) -> None:
        y_min, y_max = get_text_height(self.font, self.text, self.font_size)
        height = y_max - y_min

        if self.y_align == "top":
            self.pos.y += y_max
        elif self.y_align == "center":
            self.pos.y += y_max - height / 2
        elif self.y_align == "bottom":
            self.pos.y += y_min

        width = self.get_width()

        if self.x_align == "center":
            self.pos.x -= width / 2
        elif self.x_align == "right":
            self.pos.x -= width

    def transform_to_space(self, space: Space) -> Self:
        self.pos = chart_to_svg_coordinates(space, self.pos)
        return self

    def get_width(self) -> float:
        return get_text_width(self.font, self.text, self.font_size)

    def get_height(self) -> float:
        y_min, y_max = get_text_height(self.font, self.text, self.font_size)
        return y_max - y_min


class ElementCollection(BaseModel):
    elements: list[Element] = []

    def render(self) -> list[str]:
        return [element.render() for element in self.elements]

    def append(self, element: Element) -> None:
        self.elements.append(element)

    def extend(self, other: "ElementCollection") -> None:
        self.elements.extend(other.elements)

    def get_width(self) -> float:
        widths = [element.get_width() for element in self.elements]
        return max(widths)



from .text import Title, SubTitle, SubSubTitle, Paragraph, Notifications
from .layout import Rows, Columns, FlexRows, FlexColumns, Space
from .control import Button, EnumSelect, ListSelect
# from .chart import BarChart, LineChart


__all__ = [
    # Text
    "Title",
    "SubTitle",
    "SubSubTitle",
    "Paragraph",
    "Notifications",
    # Layout
    "Rows",
    "Columns",
    "FlexRows",
    "FlexColumns",
    "Space",
    # Control
    "Button",
    "EnumSelect",
    "ListSelect",
    # Charts
    # "BarChart",
    # "LineChart",
]



from typing import Any, Annotated
import unittest

from newsflash.callback.builder import construct_callback
from newsflash.callback.utils import process_callback_arg
from newsflash.callback.models import Callback, WidgetIO
from newsflash.widgets.chart.bar import BarChart
from newsflash.widgets import Button


class TestProcessCallbackArgs(unittest.TestCase):
    def setUp(self) -> None:
        def dummy_callback_function(
            self: Any,
            bar_chart: Annotated[BarChart, "bar-chart-id"],
            button: Annotated[Button, "button-id"],
        ):
            pass

        self.dummy_callback_function = dummy_callback_function

    def test_process_callback_arg_for_chart(self):
        result = process_callback_arg(self.dummy_callback_function, "bar_chart")

        expected = (BarChart, "bar-chart-id", WidgetIO.OUTPUT)

        self.assertEqual(result, expected)

    def test_process_callback_arg_for_button(self):
        result = process_callback_arg(self.dummy_callback_function, "button")

        expected = (Button, "button-id", WidgetIO.INPUT)

        self.assertEqual(result, expected)


class TestCallback(unittest.TestCase):
    def test_construct_callback_with_button_and_chart(self):
        def dummy_callback_function(
            self: Any,
            bar_chart: Annotated[BarChart, "bar-chart-id"],
            button: Annotated[Button, "button-id"],
        ):
            pass

        result = construct_callback(
            callback_fn=dummy_callback_function,
            endpoint_name="ABC",
            trigger_event="click",
        )

        expected = Callback(
            endpoint_name="ABC",
            trigger_event="click",
            inputs=["button-id-input"],
            targets=["bar-chart-id"],
            target_wrapper_ids="#bar-chart-id-wrapper",
        )

        self.assertEqual(result, expected)



from math import floor
from .utils import Point, Space, order_of_magnitude
from fontTools.ttLib import TTFont
from .element import ElementCollection, Text, Line


def build_x_axis(
    x_start: float,
    x_end: float,
    y_end: float,
    start_x_value: float,
    end_x_value: float,
    ticks: list[float],
    font_size: int,
    font: TTFont,
    labels: list[str] | None = None,
) -> ElementCollection:
    x_axis_height = font_size * 1.5

    x_axis_space = Space(
        top_left=Point(x=x_start, y=y_end - x_axis_height),
        bottom_right=Point(x=x_end, y=y_end),
    )

    if labels is None:
        labels = [str(tick) for tick in ticks]

    elements = ElementCollection()
    ticks = [(tick - start_x_value) / (end_x_value - start_x_value) for tick in ticks]

    for idx, tick in enumerate(ticks):
        elements.append(
            Text(
                text=labels[idx],
                font_size=font_size,
                pos=Point(x=tick, y=0),
                x_align="center",
                y_align="bottom",
                font=font,
                space=x_axis_space,
            )
        )

    return elements


def build_y_axis(
    y_start: float,
    y_end: float,
    x_start: float,
    bottom_y_value: float,
    top_y_value: float,
    ticks: list[float],
    font: TTFont,
    labels: list[str] | None = None,
) -> ElementCollection:
    y_axis_space = Space(
        top_left=Point(x=x_start, y=y_start),
        bottom_right=Point(x=x_start, y=y_end),
    )

    if labels is None:
        labels = [str(tick) for tick in ticks]

    elements = ElementCollection()
    ticks = [(tick + bottom_y_value) / top_y_value for tick in ticks]

    for idx, tick in enumerate(ticks):
        if tick >= 0 and tick <= 1:
            elements.append(
                Text(
                    text=labels[idx],
                    font_size=16,
                    font=font,
                    pos=Point(x=0, y=tick),
                    space=y_axis_space,
                )
            )

    return elements


def build_y_ticks(
    number_of_ticks: int,
    max_y_value: float,
) -> tuple[list[float], list[str], int]:
    y_step = max_y_value / (number_of_ticks - 1)
    y_step_order_of_magnitude = order_of_magnitude(y_step)

    scale_factor = pow(10, y_step_order_of_magnitude)

    y_steps = [
        round(y_step / scale_factor) * scale_factor * i for i in range(number_of_ticks)
    ]

    if max_y_value > y_steps[-1]:
        y_steps.append(round(y_step / scale_factor) * scale_factor * number_of_ticks)

    multiplier = pow(1000, floor(y_step_order_of_magnitude / 3))
    y_steps = [y_step / multiplier for y_step in y_steps]

    if y_step_order_of_magnitude < 0:
        labels = [f"{y:.{abs(y_step_order_of_magnitude)}f}" for y in y_steps]
    else:
        labels = [str(round(y)) for y in y_steps]

    return y_steps, labels, multiplier


def build_y_grid_lines(
    bottom_y_value: float,
    top_y_value: float,
    ticks: list[float],
    space: Space,
) -> ElementCollection:
    elements = ElementCollection()

    ticks = [(tick + bottom_y_value) / top_y_value for tick in ticks]

    for tick in ticks:
        if tick >= 0 and tick <= 1:
            elements.append(
                Line(
                    from_pos=Point(x=0, y=tick),
                    to_pos=Point(x=1, y=tick),
                    space=space,
                    stroke_width=2,
                    classes=["stroke-grid-lines-light", "dark:stroke-grid-lines-dark"]
                )
            )

    return elements
